MODULE ExComplexLib0;

 (*
    ExComplexLib0 - Complex number math functions.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT XM := ExMathLib0, X := ExNumbers;

TYPE
  ExComplexType * = RECORD
    re-, im- : X.ExNumType
  END;

VAR
  Ex0-, Ex1- : ExComplexType;
  ComplexToStr* : PROCEDURE (A: ExComplexType; VAR Str: ARRAY OF CHAR);


PROCEDURE Init * (VAR A : ExComplexType; re, im : X.ExNumType);
BEGIN
  A.re := re; A.im := im
END Init;


PROCEDURE Conj * (VAR x : ExComplexType);
BEGIN
  X.ExChgSign(x.im)
END Conj;


PROCEDURE PolarMag * (VAR A : X.ExNumType; B : ExComplexType);
BEGIN
  IF X.IsZero(B.im) THEN A := B.re; X.ExAbs(A)
  ELSIF X.IsZero(B.re) THEN A := B.im; X.ExAbs(A)
  ELSE
    X.ExMult(A, B.re, B.re); X.ExMult(B.re, B.im, B.im); X.ExAdd(A, B.re, A);
    XM.sqrtX(A, A)
  END;
END PolarMag;


PROCEDURE PolarAngle * (VAR A : X.ExNumType; B : ExComplexType);
BEGIN
  XM.arctan2X(A, B.re, B.im)
END PolarAngle;


PROCEDURE ToRectangular * (VAR A : ExComplexType; r, theta : X.ExNumType);
BEGIN
  XM.cosX(A.re, theta); XM.sinX(A.im, theta); 
  X.ExMult(A.re, A.re, r); X.ExMult(A.im, A.im, r);

  (* round the numbers since sin/cos are not accurate to 52 places *)
  IF A.re.Exp < 0 THEN X.ExRound(A.re, A.re.Exp+14) END;
  IF A.im.Exp < 0 THEN X.ExRound(A.im, A.im.Exp+14) END
END ToRectangular;


PROCEDURE LongRealToComplex * (x : LONGREAL; VAR xc : ExComplexType);
BEGIN
  xc.im := X.Ex0; XM.LongRealToExNum(x, xc.re)
END LongRealToComplex;


PROCEDURE ComplexToLongReal * (xc : ExComplexType) : LONGREAL;
BEGIN
  RETURN XM.ExNumToLongReal(xc.re)
END ComplexToLongReal;


PROCEDURE ExChgSign * (VAR x : ExComplexType);
BEGIN
  X.ExChgSign(x.re); X.ExChgSign(x.im)
END ExChgSign;


PROCEDURE ExAdd * (VAR A : ExComplexType; B, C : ExComplexType);
BEGIN
  (* A = B + C = (a + bi) + (c + di) = (a+c) + (b+d)i *)
  X.ExAdd(A.re, B.re, C.re); X.ExAdd(A.im, B.im, C.im);
END ExAdd;


PROCEDURE ExSub * (VAR A : ExComplexType; B, C : ExComplexType);
BEGIN
  (* A = B - C = (a + bi) - (c + di) = (a-c) + (b-d)i *)
  X.ExSub(A.re, B.re, C.re); X.ExSub(A.im, B.im, C.im);
END ExSub;


PROCEDURE ExMult * (VAR A : ExComplexType; B, C : ExComplexType);
VAR t : X.ExNumType;
BEGIN
  (* A = B * C = (a + bi) * (c + di) = (ac-bd) + (ad+bc)i *)
  X.ExMult(A.re,B.re,C.re); X.ExMult(t,B.im,C.im); X.ExSub(A.re,A.re,t);
  X.ExMult(A.im,B.re,C.im); X.ExMult(t,B.im,C.re); X.ExAdd(A.im,A.im,t)
END ExMult;


PROCEDURE ExDiv * (VAR A : ExComplexType; B, C : ExComplexType);
VAR t, div : X.ExNumType;
BEGIN
  (* A = B / C = (a + bi) / (c + di) = (ac+bd)/e + (bc-ad)/e i
             where e = c+d
   *)
  X.ExMult(div, C.re, C.re); X.ExMult(t, C.im, C.im); X.ExAdd(div, div, t);
  X.ExMult(A.re,B.re,C.re); X.ExMult(t,B.im,C.im); X.ExAdd(A.re,A.re,t);
  X.ExMult(A.im,B.im,C.re); X.ExMult(t,B.re,C.im); X.ExSub(A.im,A.im,t);
  X.ExDiv(A.re, A.re, div); X.ExDiv(A.im, A.im, div)
END ExDiv;


PROCEDURE xtoi * (VAR Result : ExComplexType; x : ExComplexType; i : LONGINT);
VAR
  Y : ExComplexType;
  negative : BOOLEAN;
BEGIN
  Y := Ex1;
  negative := i < 0;
  i := ABS(i);
  LOOP
    IF ODD(i) THEN ExMult(Y, Y, x) END;
    i := i DIV 2;
    IF i = 0 THEN EXIT END;
    ExMult(x, x, x);
  END;
  IF negative THEN
    ExDiv(Result, Ex1, Y);
  ELSE
    Result := Y;
  END;
END xtoi;


PROCEDURE Root * (VAR Result : ExComplexType; x : ExComplexType; i : LONGINT);
VAR r, theta : X.ExNumType; izero: BOOLEAN;
BEGIN
  izero:=X.IsZero(x.im);
  IF izero & (x.re.Sign = X.positive) THEN
    XM.Root(Result.re, x.re, i); Result.im := X.Ex0
  ELSIF izero & ODD(i) THEN
    X.ExAbs(x.re); XM.Root(Result.re, x.re, i); 
    X.ExChgSign(Result.re); Result.im := X.Ex0
  ELSE PolarMag(r, x); PolarAngle(theta, x); X.ExNumb(i, 0, 0, x.re);
    XM.Root(r, r, i); X.ExDiv(theta, theta, x.re);
    ToRectangular(Result, r, theta)
  END
END Root;


PROCEDURE FromRadians*(VAR radianAngle : ExComplexType);
(* Convert a radian measure into current angle measure *)
BEGIN
  XM.FromRadians(radianAngle.re); XM.FromRadians(radianAngle.im)
END FromRadians;


PROCEDURE ToRadians*(VAR radianAngle : ExComplexType);
(* Convert an angle measure into radians *)
BEGIN
  XM.ToRadians(radianAngle.re); XM.ToRadians(radianAngle.im)
END ToRadians;


PROCEDURE sqrtX*(VAR Result : ExComplexType; x : ExComplexType);
BEGIN
  Root(Result, x, 2);
END sqrtX;


PROCEDURE lnX*(VAR Result : ExComplexType; x : ExComplexType);
VAR half : X.ExNumType; pos : BOOLEAN;
BEGIN
  IF X.IsZero(x.im) THEN pos := x.re.Sign = X.positive;
    X.ExAbs(x.re); XM.lnX(Result.re, x.re);
    IF pos THEN Result.im := X.Ex0 ELSE Result.im := X.pi END
  ELSIF X.IsZero(x.re) THEN pos := x.im.Sign = X.positive;
    X.ExAbs(x.im); XM.lnX(Result.re, x.im); X.ExNumb(0, 5, 0, half);
    X.ExMult(Result.im, half, X.pi);  (* pi/2 *)
    IF ~pos THEN X.ExChgSign(Result.im) END
  ELSE
    PolarMag(Result.re, x); PolarAngle(Result.im, x);
    XM.lnX(Result.re, Result.re)
  END
END lnX;


PROCEDURE logX*(VAR Result : ExComplexType; x : ExComplexType);
BEGIN
  lnX(x, x);
  X.ExMult(Result.re, X.lne, x.re); X.ExMult(Result.im, X.lne, x.im)
END logX;


PROCEDURE factorialX*(VAR Result : ExComplexType; n : LONGINT);
BEGIN
  XM.factorialX(Result.re, n); Result.im := X.Ex0
END factorialX;


PROCEDURE expX*(VAR Result : ExComplexType; x : ExComplexType);
BEGIN
  XM.expX(x.re, x.re); ToRectangular(Result, x.re, x.im)
END expX;


PROCEDURE powerX*(VAR Result : ExComplexType; x, y : ExComplexType);
VAR p : LONGREAL; t : ExComplexType;
BEGIN
  IF X.IsZero(y.im) & (y.re.Sign = X.positive) THEN (* just real powers *)
    IF X.IsZero(x.im) & (x.re.Sign = X.positive) THEN (* real numbers *)
      XM.powerX(Result.re, x.re, y.re); Result.im := X.Ex0
    ELSE p := ComplexToLongReal(y); t := y; X.ExFrac(t.re);
      IF (ABS(p) < MAX(LONGINT)) & X.IsZero(t.re) THEN xtoi(Result, x, ENTIER(p))
      ELSE lnX(Result, x); ExMult(x, Result, y); expX(Result, x)
      END
    END
  ELSE
    lnX(Result, x); ExMult(x, Result, y); expX(Result, x)
  END
END powerX;


PROCEDURE rootX*(VAR Result : ExComplexType; x, y : ExComplexType);
VAR r : LONGREAL; t : ExComplexType;
BEGIN
  IF X.IsZero(y.im) & (y.re.Sign = X.positive) THEN (* just real roots *)
    IF X.IsZero(x.im) & (x.re.Sign = X.positive) THEN (* real numbers *)
      XM.rootX(Result.re, x.re, y.re); Result.im := X.Ex0
    ELSE r := ComplexToLongReal(y); t := y; X.ExFrac(t.re);
      IF (ABS(r) < MAX(LONGINT)) & X.IsZero(t.re) THEN Root(Result, x, ENTIER(r))
      ELSE lnX(Result, x); ExDiv(x, Result, y); expX(Result, x)
      END
    END
  ELSE lnX(Result, x); ExDiv(x, Result, y); expX(Result, x)
  END
END rootX;


PROCEDURE sinX*(VAR Result : ExComplexType; x : ExComplexType);
VAR t : X.ExNumType;
BEGIN
  XM.sinX(t, x.re); XM.coshX(Result.re, x.im); X.ExMult(Result.re, t, Result.re);
  XM.cosX(t, x.re); XM.sinhX(Result.im, x.im); X.ExMult(Result.im, t, Result.im)
END sinX;


PROCEDURE cosX*(VAR Result : ExComplexType; x : ExComplexType);
VAR t : X.ExNumType;
BEGIN
  XM.cosX(t, x.re); XM.coshX(Result.re, x.im); X.ExMult(Result.re, t, Result.re);
  XM.sinX(t, x.re); XM.sinhX(Result.im, x.im); X.ExMult(Result.im, t, Result.im);
  X.ExChgSign(Result.im)  (* negate *)
END cosX;


PROCEDURE tanX*(VAR Result : ExComplexType; x : ExComplexType);
VAR TWO, s, c, d: X.ExNumType;
BEGIN
  X.ExNumb(2, 0, 0, TWO);
  X.ExMult(x.re, TWO, x.re); X.ExMult(x.im, TWO, x.im);
  XM.sinX(s, x.re); XM.cosX(c, x.re);
  XM.coshX(d, x.im); X.ExAdd(d, d, c);
  X.ExDiv(Result.re, s, d); 
  XM.sinhX(Result.im, x.im); 
  X.ExDiv(Result.im, Result.im, d)
END tanX;


PROCEDURE CalcAlphaBeta (z : ExComplexType; VAR a, b: X.ExNumType);
VAR x, x2, y, r, t, HALF: X.ExNumType;
BEGIN
  X.ExNumb(0, 5, 0, HALF);
  X.ExAdd(x, z.re, X.Ex1); X.ExMult(x, x, x); X.ExMult(y, z.im, z.im);
  X.ExSub(x2, z.re, X.Ex1); X.ExMult(x2, x2, x2);
  X.ExAdd(t, x, y); XM.sqrtX(r, t);
  X.ExAdd(t, x2, y); XM.sqrtX(t, t);
  X.ExAdd(a, r, t); X.ExMult(a, a, HALF);
  X.ExSub(b, r, t); X.ExMult(b, b, HALF)
END CalcAlphaBeta;


PROCEDURE arcsinX*(VAR Result : ExComplexType; x : ExComplexType);
VAR a, b: X.ExNumType;
BEGIN
  CalcAlphaBeta(x, a, b);
  XM.arcsinX(Result.re, b);
  X.ExMult(b, a, a); X.ExSub(b, b, X.Ex1); XM.sqrtX(b, b);
  X.ExAdd(b, a, b); XM.lnX(Result.im, b)
END arcsinX;


PROCEDURE arccosX*(VAR Result : ExComplexType; x : ExComplexType);
VAR a, b: X.ExNumType;
BEGIN
  CalcAlphaBeta(x, a, b);
  XM.arccosX(Result.re, b);
  X.ExMult(b, a, a); X.ExSub(b, b, X.Ex1); XM.sqrtX(b, b);
  X.ExAdd(b, a, b); XM.lnX(Result.im, b); X.ExChgSign(Result.im)
END arccosX;


PROCEDURE arctanX*(VAR Result : ExComplexType; z : ExComplexType);
VAR x, x2, y2, y, t, yp, TWO, HALF, QUARTER: X.ExNumType;
BEGIN
  X.ExNumb(2, 0, 0, TWO); X.ExNumb(0, 5, 0, HALF); X.ExNumb(0, 25, 0, QUARTER);
  X.ExMult(x, TWO, z.re); X.ExAdd(y, z.im, X.Ex1); X.ExMult(y, y, y);
  X.ExSub(yp, z.im, X.Ex1); X.ExMult(yp, yp, yp);
  X.ExMult(x2, z.re, z.re); X.ExMult(y2, z.im, z.im);
  X.ExSub(t, X.Ex1, x2); X.ExSub(t, t, y2); X.ExDiv(t, x, t);
  XM.arctanX(t, t); X.ExMult(Result.re, HALF, t);
  X.ExAdd(t, x2, y); X.ExAdd(y, x2, yp); X.ExDiv(t, t, y);
  XM.lnX(t, t); X.ExMult(Result.im, QUARTER, t)
END arctanX;


PROCEDURE sinhX*(VAR Result : ExComplexType; x : ExComplexType);
VAR t : X.ExNumType;
BEGIN
  XM.sinhX(t, x.re); XM.coshX(Result.re, x.im); X.ExMult(Result.re, t, Result.re);
  XM.coshX(t, x.re); XM.sinhX(Result.im, x.im); X.ExMult(Result.im, t, Result.im)
END sinhX;


PROCEDURE coshX*(VAR Result : ExComplexType; x : ExComplexType);
VAR t : X.ExNumType;
BEGIN
  XM.coshX(t, x.re); XM.coshX(Result.re, x.im); X.ExMult(Result.re, t, Result.re);
  XM.sinhX(t, x.re); XM.sinhX(Result.im, x.im); X.ExMult(Result.im, t, Result.im)
END coshX;


PROCEDURE tanhX*(VAR Result : ExComplexType; x : ExComplexType);
BEGIN
  sinhX(Result, x); coshX(x, x); ExDiv(Result, Result, x)
END tanhX;


PROCEDURE arcsinhX*(VAR Result : ExComplexType; x : ExComplexType);
VAR
  Temp : ExComplexType;
BEGIN
  (* Result = ln(x + sqrt(x*x + 1)) *)
  ExMult(Temp, x, x); ExAdd(Temp, Temp, Ex1); sqrtX(Temp, Temp);
  ExAdd(Temp, x, Temp); lnX(Result, Temp);
END arcsinhX;


PROCEDURE arccoshX*(VAR Result : ExComplexType; x : ExComplexType);
VAR
  Temp : ExComplexType;
BEGIN
  (* Result = ln(x + sqrt(x*x - 1)) *)
  ExMult(Temp, x, x); ExSub(Temp, Temp, Ex1); sqrtX(Temp, Temp);
  ExAdd(Temp, x, Temp); lnX(Result, Temp);
END arccoshX;


PROCEDURE arctanhX*(VAR Result : ExComplexType; x : ExComplexType);
VAR
  Temp, Temp2 : ExComplexType;
  Half : X.ExNumType;
BEGIN
  (* Result = ln((1 + x) / (1 - x)) / 2 *)
  ExAdd(Temp, Ex1, x); ExSub(Temp2, Ex1, x); ExDiv(Temp, Temp, Temp2);
  IF X.ExStatus=X.Okay THEN
    lnX(Result, Temp);
    X.ExNumb(0, 5, 0, Half); Init(Temp, Half, X.Ex0); ExMult(Result, Result, Temp)
  END
END arctanhX;


BEGIN
  Ex1.re := X.Ex1; Ex1.im := X.Ex0;
  Ex0.re := X.Ex0; Ex0.im := X.Ex0
END ExComplexLib0.
