MODULE UnixPlotLibrary;

 (*
    UnixPlotLibrary - Plotting facility.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT m := LRealMath, str := Strings, X := ExNumbers;

CONST
  BackColour = 0;
  Colour1    = 1;
  Colour2    = 2;
  Colour3    = 3;

  CharHeight = 8;
  CharWidth  = 8;

  (* PlotKindType *)
  Normal * = 0;
  LogX   * = 1;
  LogY   * = 2;
  Log    * = 3;
  Line   * = 4;
  Bar    * = 5;

  (* PositionType *)
  UpperLeft   * = 0;
  UpperMiddle * = 1;
  UpperRight  * = 2;
  LowerLeft   * = 3;
  LowerMiddle * = 4;
  LowerRight  * = 5;

  (* ClassType *)
  Left   * = 0;
  Right  * = 1;
  Top    * = 2;
  Bottom * = 3;

TYPE
  PlotFunction * = PROCEDURE(x : LONGREAL) : LONGREAL;
  LabelProc    * = PROCEDURE(x : INTEGER; VAR labels : ARRAY OF CHAR);

  PlotKindType * = SHORTINT;

  LineType     * = ARRAY 80 OF CHAR;
  PositionType * = SHORTINT;

  ClassType * = SHORTINT;
  ClassSet  * = SET;
  PlotSet   * = SET;

CONST
  XLog = {LogX, Log};
  YLog = {LogY, Log};

TYPE
  PlotType * = POINTER TO PlotRecord;
  PlotRecord * =
    RECORD
      (*
      Window                 : i.WindowPtr;
      Drawing                : g.RastPortPtr;
      *)
      PlotIs                 : PlotKindType;
      GridColour             : INTEGER;
      TextColour             : INTEGER;
      PlotColour             : INTEGER;
      UserLabels             : BOOLEAN;
      GetLabel               : LabelProc;(* label routine *)
      ScatterPlot            : BOOLEAN;  (* scatter plots *)
      ScatterChar            : ARRAY 2 OF CHAR;
      ZeroLineOn             : BOOLEAN;  (* line on/off   *)
      Minx-, Miny-, Maxx-, Maxy- : LONGREAL; (* real coords.  *)
      GMnx, GMny, GMxx, GMxy : INTEGER;  (* screen outline*)
      Scrw, Scrh             : INTEGER;  (* & dimensions. *)
      Divx, Divy,                        (* major divs.   *)
      SubDivx, SubDivy,                  (* minor divs.   *)
      Decx, Decy             : INTEGER;  (* decimal places*)
      xscale, yscale         : LONGREAL; (* x/y scaling   *)
      xoff, yoff             : INTEGER;  (* plot offsets  *)
      GridDrawn              : BOOLEAN;  (* draw once     *)
      CopyOfAPlot            : BOOLEAN;  (* not original  *)
    END;

(*-------------------------------------------------------*)
(* The following are a number of utility routines used   *)
(* only within this module to facilitate the graphing.   *)


CONST
  ZERO  = 0.0D0;
  ONE   = 1.0D0;
  TEN   = 10.0D0;
  TENTH = 0.1D0;


PROCEDURE PowerOfTen(power:INTEGER): LONGREAL;    (*10^power*)
VAR
  result : LONGREAL;
BEGIN
  result := ONE;
  IF power > 0 THEN
    REPEAT
      result := result * TEN;
      DEC(power);
    UNTIL power=0;
  ELSIF power < 0 THEN
    REPEAT
      result := result / TEN;
      INC(power);
    UNTIL power=0;
  END;
  RETURN result;
END PowerOfTen;


PROCEDURE Trunc (x : LONGREAL) : LONGINT;
VAR r: LONGINT;
BEGIN
  r:=ENTIER(x);
  IF x >= ZERO THEN
    IF r>x THEN RETURN r-1 ELSE RETURN r END (* floor *)
  ELSE
    IF r<x THEN RETURN r+1 ELSE RETURN r END (* ceiling *)
  END
END Trunc;


PROCEDURE GetExp(num:LONGREAL): INTEGER;
BEGIN
  IF num = ZERO THEN RETURN 0
  ELSE RETURN SHORT(Trunc(m.log(ABS(num), 10)))
  END
END GetExp;


PROCEDURE RealToString * (num:LONGREAL;
                          VAR str:ARRAY OF CHAR;
                          minchars:INTEGER;
                          fracchars:INTEGER;
                          pad:CHAR): BOOLEAN;
CONST
  RAcc = 15; (* digits of accuracy in REAL we can count on *)
VAR
  D,s,exp : INTEGER;
  buffer  : ARRAY 80 OF CHAR;
  pos,i,j : INTEGER;
  useExp  : BOOLEAN;

  PROCEDURE PutCh(c:CHAR);
  BEGIN
    buffer[pos]:=c;
    INC(pos);
  END PutCh;

  PROCEDURE PutNumDig(c:LONGINT);
  BEGIN
    buffer[pos]:=CHR(c+ORD("0"));
    INC(pos);
  END PutNumDig;

  PROCEDURE PutExp(exp:INTEGER);
  BEGIN
    PutCh('E');
    IF exp<0 THEN
      PutCh('-');
      exp:=-exp;
    END;

    IF exp>10 THEN
      PutNumDig((exp MOD 100) DIV 10);
    END;

    PutNumDig(exp MOD 10);
  END PutExp;

BEGIN
  pos:=0;

  D := ABS(fracchars);
  useExp:= (fracchars < 0);

  IF num < ZERO THEN
    PutCh("-");
    num := -num;
  END;

  IF ABS(num) < 1.0E-5 THEN
    exp := 0;
    num := ZERO;
  ELSE
    (* round to D digits to the right of the decimal *)
    IF D > RAcc THEN
      D := RAcc;
    END;
    num:=num+PowerOfTen(-D)*0.5;

    (* normalize number *)
    exp := GetExp(num);
    num := num / PowerOfTen(exp);

    (* check if rounding changed first digit of mantissa *)
    (* i.e., 9.99999 changed to 10.00000 *)
    IF num >= TEN THEN
      num := num / TEN;
      INC(exp);
    END;

    (* special case when ABS(num) < 1.0 *)
    IF num < 1.0 THEN
      num := num * TEN;
      DEC(exp);
    END;

    (* check if this number can be represented
       in ordinary form *)
    IF ABS(exp) > RAcc THEN
      useExp := TRUE;
    END;
  END;

  IF useExp THEN
    PutNumDig(Trunc(num));
    PutCh(".");

    j:=D;
    WHILE j>0 DO
      num := (num - Trunc(num)) * TEN;
      PutNumDig(Trunc(num));
      DEC(j);
    END;

    PutExp(exp);

  ELSE
    s := exp;  (* begin with digits to left of decimal *)
    IF exp < 0 THEN
      s := 0;  (* begin at zero to left of decimal *)
    END;

    i := 0;
    FOR j := s TO (-D) BY -1 DO
      IF (i > RAcc) OR (exp < 0) THEN
      (* use dummy zeros to pad right *)
        PutCh("0");
      ELSE
        PutNumDig(Trunc(num));
        num := (num - Trunc(num)) * TEN;
      END;
      INC(i);
      INC(exp);
      IF j = 0 THEN
        PutCh(".");
      END;
    END;
  END;

  PutCh(0X);

  IF (minchars>LEN(str)-1) OR (pos>LEN(str)-1) THEN
    RETURN FALSE;
  END;

  i:=0;                                 (* put padding *)
  WHILE i+pos<=minchars DO
    str[i]:=pad;
    INC(i);
  END;

  j:=-1;
  (* copy data into destination *)
  REPEAT
    INC(j);
    str[i]:=buffer[j];
    INC(i);
  UNTIL buffer[j]=0X;

  IF (j > 0) & (str[j-1] = '.') THEN
    (* remove trailing '.' *)
    str[j-1] := 0X;
  END;

  RETURN TRUE;
END RealToString;


PROCEDURE Round(In : LONGREAL) : INTEGER;
BEGIN
  IF In > MAX(INTEGER) THEN RETURN MAX(INTEGER)
  ELSIF In < MIN(INTEGER) THEN RETURN MIN(INTEGER)
  ELSE RETURN SHORT(ENTIER(In))
  END
END Round;


PROCEDURE TenTo(x : LONGREAL) : LONGREAL;
BEGIN
  RETURN m.power(10, x);
END TenTo;


PROCEDURE LogOf(x : LONGREAL) : LONGREAL;
BEGIN
  IF x <= 0.0 THEN (* capture illegal values *)
    RETURN 0.0;    (* and return log(1)      *)
  END;
  RETURN m.log(x, 10);
END LogOf;


PROCEDURE xyToCoords(p          : PlotType;
                     x, y       : LONGREAL;
                     VAR xG, yG : INTEGER);
(* Convert floating point world coordinates to graph space
   coordinates. *)
BEGIN
  xG := Round((x-p.Minx)*(p.GMxx-p.GMnx)/(p.Maxx-p.Minx))+p.GMnx;
  yG := p.GMxy-Round((y-p.Miny)*(p.GMxy-p.GMny)/(p.Maxy-p.Miny))
END xyToCoords;


PROCEDURE CoordsToxy(p        : PlotType;
                     xG, yG   : INTEGER;
                     VAR x, y : LONGREAL);
(* Convert graph space coordinates to floating point world
   coordinates. *)
BEGIN
  x := (xG-p.GMnx)*(p.Maxx-p.Minx)/(p.GMxx-p.GMnx)+p.Minx;
  y := (p.GMxy-yG)*(p.Maxy-p.Miny)/(p.GMxy-p.GMny)+p.Miny
END CoordsToxy;


PROCEDURE OffsetPoint(p : PlotType; VAR x, y : INTEGER);
(* Offset the graph points by the plot amounts *)
BEGIN
  INC(x, p.xoff);
  INC(y, p.yoff);
END OffsetPoint;


PROCEDURE DrawLine(p                  : PlotType;
                   Colour             : INTEGER;
                   Gx1, Gy1, Gx2, Gy2 : INTEGER);
(* Line clipping algorithm : two points which make up a
   line are passed (Gx1, Gy1) and (Gx2, Gy2) to be
   clipped within the Plot space.  NOTE: The Amiga screen
   coordinates have y = 0 at the top of the screen.

   From an algorithm published in "Microcomputer Graphics
   for the IBM PC" by Roy E. Myers. *)
VAR
  x, y    : ARRAY 4 OF INTEGER;
  xl, xh  : INTEGER;
  yl, yh  : INTEGER;
  Bound   : ARRAY 4 OF ClassSet;
  Discard : BOOLEAN;
  i       : INTEGER;

  PROCEDURE Classify (VAR Boundary : ClassSet;
                      x, y         : INTEGER);
  BEGIN
    Boundary := {};           (* assume inside bounds  *)
    IF x < xl THEN
      INCL(Boundary, Left);   (* outside left bounds   *)
    ELSIF x > xh THEN
      INCL(Boundary, Right);  (* outside right bounds  *)
    END;
    IF y > yh THEN
      INCL(Boundary, Bottom); (* outside bottom bounds *)
    ELSIF y < yl THEN
      INCL(Boundary, Top);    (* outside top bounds    *)
    END;
  END Classify;

  PROCEDURE Clip (Boundary : ClassSet);
  VAR
    dy, dx : LONGINT;
  BEGIN
    (* NOTE: 32-bit calculations used to avoid range errors *)
    dy := y[2] - y[1];
    dx := x[2] - x[1];
    IF Left IN Boundary THEN
      x[i] := xl;
      y[i] := SHORT(y[1] + dy * (LONG(xl) - x[1]) DIV dx);
    ELSIF Right IN Boundary THEN
      x[i] := xh;
      y[i] := SHORT(y[1] + dy * (LONG(xh) - x[1]) DIV dx);
    ELSIF Top IN Boundary THEN
      x[i] := SHORT(x[1] + dx * (LONG(yl) - y[1]) DIV dy);
      y[i] := yl;
    ELSIF Bottom IN Boundary THEN
      x[i] := SHORT(x[1] + dx * (LONG(yh) - y[1]) DIV dy);
      y[i] := yh;
    END
  END Clip;

  PROCEDURE PlotLine (x1, y1, x2, y2 : INTEGER);
  (* Draw a clipped line to the screen. *)
  BEGIN
    (*
    g.SetAPen(p.Drawing, Colour);    (* set line colour *)
    IF p.ScatterPlot THEN
      (* move to the character centre and place char *)
      g.Move(p.Drawing, x1-HalfWidth, y1+HalfHeight);
      g.Text(p.Drawing, p.ScatterChar, 1);
    ELSE
      (* move to 1st pt and draw to 2nd point *)
      g.Move(p.Drawing, x1, y1);
      g.Draw(p.Drawing, x2, y2);
    END
    *)
  END PlotLine;

BEGIN
  (* Initialize clipping routines *)
  x[1] := Gx1;  y[1] := Gy1;
  x[2] := Gx2;  y[2] := Gy2;
  xl := p.GMnx; xh := p.GMxx;
  yl := p.GMny; yh := p.GMxy;

  (* Classify the endpoints *)
  Classify(Bound[1], x[1], y[1]);
  Classify(Bound[2], x[2], y[2]);

  (* Perform the line clipping *)
  IF Bound[1] + Bound[2] = {} THEN
    (* Within plot area -- just draw the line *)
    PlotLine(x[1], y[1], x[2], y[2]);
  ELSE
    (* Just output a clipped line segment *)
    i := 1;
    REPEAT
      Discard := Bound[1] * Bound[2] # {};
      IF ~Discard THEN
        IF Bound[i] # {} THEN
          (* Handle the line clipping *)
          Clip(Bound[i]);
          Classify(Bound[i], x[i], y[i]);
        ELSE
          INC(i);
        END;
      END;
    UNTIL (i > 2) OR Discard;
    IF ~Discard THEN
      PlotLine(x[1], y[1], x[2], y[2]);  (* Clipped line *)
    END;
  END;
END DrawLine;


PROCEDURE IntPlotLine(Plot           : PlotType;
                      Colour         : INTEGER;
                      x1, y1, x2, y2 : LONGREAL);
(* Draw a line between points (x1,y1) and (x2,y2).

   This routine accepts log(x) values and displays them
   properly during log plots. *)
VAR
  Gx1, Gy1, Gx2, Gy2 : INTEGER;
BEGIN
  xyToCoords(Plot, x1, y1, Gx1, Gy1);  (* to pixel coords *)
  xyToCoords(Plot, x2, y2, Gx2, Gy2);
  OffsetPoint(Plot, Gx1, Gy1);         (* offset plot *)
  OffsetPoint(Plot, Gx2, Gy2);
  DrawLine(Plot, Colour, Gx1, Gy1, Gx2, Gy2);
END IntPlotLine;


PROCEDURE PlotLine(Plot           : PlotType;
                   Colour         : INTEGER;
                   x1, y1, x2, y2 : LONGREAL);
(* Draw a line between points (x1,y1) and (x2,y2).

   This routine converts points to logarithmic points
   during log plots to make external appearances
   identical to the caller of this routine. *)
BEGIN
  CASE Plot.PlotIs OF
    LogX :
      IntPlotLine(Plot, Colour, LogOf(x1), y1, LogOf(x2), y2);
      |
    LogY :
      IntPlotLine(Plot, Colour, x1, LogOf(y1), x2, LogOf(y2));
      |
    Log :
      IntPlotLine(Plot, Colour, LogOf(x1), LogOf(y1), LogOf(x2), LogOf(y2));
      |
    ELSE
      IntPlotLine(Plot, Colour, x1, y1, x2, y2);
  END;
END PlotLine;


PROCEDURE PlotBar(Plot   : PlotType;
                  Colour : INTEGER;
                  width  : LONGREAL;
                  x, y   : LONGREAL);
(* Draw a bar centered around x of height y and width
   across. *)
VAR
  Half : LONGREAL;
  Gxl, Gyl, Gxu, Gyu : INTEGER;
BEGIN
  Half := ABS(width) / 2.0;
  IF y >= 0.0 THEN
    xyToCoords(Plot, x-Half, y, Gxl, Gyl);
    xyToCoords(Plot, x+Half, 0.0, Gxu, Gyu);
  ELSE
    xyToCoords(Plot, x-Half, 0.0, Gxl, Gyl);
    xyToCoords(Plot, x+Half, y, Gxu, Gyu);
  END;

  IF Gyl # Gyu THEN
    OffsetPoint(Plot, Gxl, Gyl);    (* offset bar *)
    OffsetPoint(Plot, Gxu, Gyu);
    (*
    g.SetAPen(Plot.Drawing, Colour);
    g.RectFill(Plot.Drawing, Gxl, Gyl, Gxu, Gyu);
    *)
  END;
END PlotBar;


PROCEDURE LabelX * (p     : PlotType;
                    Label : ARRAY OF CHAR;
                    x, y  : INTEGER);
(* Place text horizontally on the plot *)
BEGIN
  (* Change bounds of screen to allow text outside plot *)
  (*
  IF (x >= 2)  & (x <= p.Window.width-1) &
     (y >= 12) & (y <= p.Window.height-1) THEN
    g.SetAPen(p.Drawing, p.TextColour); (* set text colour *)
    g.Move(p.Drawing, x, y);            (* set text position *)
    g.Text(p.Drawing, Label, str.Length(Label));
  END
  *)
END LabelX;


PROCEDURE LabelY * (Plot  : PlotType;
                    Label : ARRAY OF CHAR;
                    x, y  : INTEGER);
(* Place text vertically on the plot *)
VAR
  ChCnt : INTEGER;
  Str   : ARRAY 2 OF CHAR;
BEGIN
  Str[1] := 0X;
  FOR ChCnt := 1 TO SHORT(str.Length(Label)) DO
    Str[0] := Label[ChCnt-1];
    LabelX(Plot, Str, x, y);
    INC(y, CharHeight);
  END;
END LabelY;


PROCEDURE PlotXLabel(p    : PlotType;
                     int  : INTEGER;
                     x, y : LONGREAL);
(* Place a label on the x-axis *)
VAR
  Gx, Gy : INTEGER;
  xlog   : LONGREAL;
  s      : ARRAY 40 OF CHAR;
BEGIN
  (* compensate for log scales along x-axis *)
  IF p.PlotIs IN XLog THEN
    xlog := TenTo(x);
  ELSE
    xlog := x;
  END;

  (* produce the x-axis labels *)
  IF p.UserLabels THEN
    p.GetLabel(int, s);   (* call-back to get label   *)
  ELSIF ~RealToString(xlog, s, 2, p.Decx, " ") THEN
    s := "***";         (* number conversion failed *)
  END;

  (* label the axis *)
  xyToCoords(p, x, y, Gx, Gy);
  DEC(Gx, str.Length(s) * CharWidth DIV 2);
  INC(Gy, 5 * CharHeight DIV 4);
  LabelX(p, s, Gx, Gy);
END PlotXLabel;


PROCEDURE PlotYLabel(p    : PlotType;
                     x, y : LONGREAL);
(* Place a label on the y-axis *)
VAR
  Gx, Gy : INTEGER;
  ylog   : LONGREAL;
  s      : ARRAY 40 OF CHAR;
BEGIN
  (* compensate for log scales along y-axis *)
  IF p.PlotIs IN YLog THEN
    ylog := TenTo(y);
  ELSE
    ylog := y;
  END;

  (* produce the y-axis labels *)
  IF ~RealToString(ylog, s, 2, p.Decy, " ") THEN
    s := "***";
  END;
  xyToCoords(p, x, y, Gx, Gy);
  DEC(Gx, (str.Length(s) + 1) * CharWidth);
  INC(Gy, CharHeight DIV 4);
  LabelX(p, s, Gx, Gy);
END PlotYLabel;


PROCEDURE ClearPlot * (p : PlotType);
(* Clear the plot area *)
VAR
  xl, xu, yl, yu : INTEGER;
  (* w : i.WindowPtr; *)
BEGIN
  (*
  g.SetAPen(p.Drawing, BackColour); (* background colour *)
  w := p.Window;
  xl := w.borderLeft+4;
  xu := xl+p.Scrw-16;
  yl := w.borderTop+4;
  yu := yl+p.Scrh-22;
  g.RectFill(p.Drawing, xl, yl, xu, yu);
  *)
  p.GridDrawn := FALSE;
END ClearPlot;


PROCEDURE SetPlotOffset * (Plot       : PlotType;
                           xoff, yoff : INTEGER);
(* Alter the current plot base position -- useful for
   realigning bar charts or giving three dimensional
   looking graphs *)
BEGIN
  Plot.xoff := xoff;
  Plot.yoff := yoff;
END SetPlotOffset;


PROCEDURE DrawGrid(p : PlotType);
(* Create the graph grid. *)
CONST
  DashLine = 05555H;
  NormLine = -1;
VAR
  vCnt, hCnt : INTEGER;
  oldx, oldy : INTEGER;
  x, y, xint, yint : LONGREAL;
  ScatterState : BOOLEAN;

  PROCEDURE PlotMinorX;
  VAR
    Cnt : INTEGER;
    xm, xi : LONGREAL;
  BEGIN
    IF p.SubDivx > 1 THEN
      (* Change to a dashed line *)
      (* g.SetDrPt(p.Drawing, DashLine); *)

      (* Output the minor grid lines *)
      IF p.PlotIs IN XLog THEN
        xi := TenTo(xint) / (p.SubDivx+1);
        FOR Cnt := 2 TO p.SubDivx DO
          xm := x + LogOf(Cnt * xi);
          IntPlotLine(p, p.GridColour, xm, p.Miny, xm, p.Maxy);
        END;
      ELSE
        xi := xint / p.SubDivx;
        FOR Cnt := 1 TO p.SubDivx-1 DO
          xm := x + Cnt * xi;
          IntPlotLine(p, p.GridColour, xm, p.Miny, xm, p.Maxy);
        END;
      END;

      (* Back to a normal line *)
      (* g.SetDrPt(p.Drawing, NormLine); *)
    END;
  END PlotMinorX;

  PROCEDURE PlotMinorY;
  VAR
    Cnt : INTEGER;
    ym, yi : LONGREAL;
  BEGIN
    IF p.SubDivy > 1 THEN
      (* Change to a dashed line *)
      (* g.SetDrPt(p.Drawing, DashLine); *)

      (* Output the minor grid lines *)
      IF p.PlotIs IN YLog THEN
        yi := TenTo(yint) / (p.SubDivy+1);
        FOR Cnt := 2 TO p.SubDivy DO
          ym := y + LogOf(Cnt * yi);
          IntPlotLine(p, p.GridColour, p.Minx, ym, p.Maxx, ym);
        END;
      ELSE
        yi := yint / p.SubDivy;
        FOR Cnt := 1 TO p.SubDivy-1 DO
          ym := y + Cnt * yi;
          IntPlotLine(p, p.GridColour, p.Minx, ym, p.Maxx, ym);
        END;
      END;

      (* Back to a normal line *)
      (* g.SetDrPt(p.Drawing, NormLine); *)
    END;
  END PlotMinorY;

BEGIN
  (* IF p.GridDrawn THEN RETURN END; *)

  (* Draw the vertical grid lines or label bars *)
  ScatterState := p.ScatterPlot;
  p.ScatterPlot := FALSE;      (* draw lines!     *)
  ClearPlot(p);                (* clear plot area *)
  oldx := p.xoff;
  oldy := p.yoff;
  SetPlotOffset(p, 0, 0);      (* no grid offset  *)
  (* g.SetDrMd(p.Drawing, g.jam1); *)
  IF p.Divx > 0 THEN
    xint := (p.Maxx - p.Minx) / p.Divx;
    FOR vCnt := 0 TO p.Divx DO
      x := p.Minx + vCnt * xint;
      PlotMinorX;
      IntPlotLine(p, p.GridColour, x, p.Miny, x, p.Maxy);
      PlotXLabel(p, vCnt, x, p.Miny);
    END;
  END;

  (* Draw the horizontal grid lines *)
  IF p.Divy > 0 THEN
    yint := (p.Maxy - p.Miny) / p.Divy;
    FOR hCnt := 0 TO p.Divy DO
      y := p.Miny + hCnt * yint;
      PlotMinorY;
      IntPlotLine(p, p.GridColour, p.Minx, y, p.Maxx, y);
      PlotYLabel(p, p.Minx, y);
    END;
  END;
  SetPlotOffset(p, oldx, oldy);   (* user offset *)
  (* g.SetDrMd(p.Drawing, g.jam2); *)
  p.GridDrawn := TRUE;
  p.ScatterPlot := ScatterState;
END DrawGrid;


(*-------------------------------------------------------*)
(* The main graphing routines.                           *)

PROCEDURE Empty * () : PlotType;
(* Used to initialize or pass empty plot parameters *)
BEGIN
  RETURN NIL;
END Empty;


PROCEDURE SetPlotLimits * (p                      : PlotType;
                           xMin, xMax, yMin, yMax : LONGREAL);
(* Alter the current plot limits -- useful for overlaying
   plots with different minima/maxima *)
BEGIN
  p.Miny := yMin;  p.Maxy := yMax;
  p.Minx := xMin;  p.Maxx := xMax;
  IF p.PlotIs IN XLog THEN
    p.Minx := LogOf(xMin);
    IF xMax <= 0.0 THEN (* make a log multiple of xMin *)
      p.Maxx := p.Minx + p.Divx;
    ELSE
      p.Maxx := LogOf(xMax);
    END;
  END;
  IF p.PlotIs IN YLog THEN
    p.Miny := LogOf(yMin);
    IF yMax <= 0.0 THEN (* make a log multiple of yMin *)
      p.Maxy := p.Miny + p.Divy;
    ELSE
      p.Maxy := LogOf(yMax);
    END;
  END;
END SetPlotLimits;


PROCEDURE SetPlotScale * (Plot           : PlotType;
                          xscale, yscale : LONGREAL);
(* Alter the plot scaling along x- and y-axes *)
BEGIN
  Plot.xscale := ABS(xscale);
  Plot.yscale := ABS(yscale);

  (* limit to 1/10 reduction *)
  IF Plot.xscale < 0.1 THEN
    Plot.xscale := 0.1;
  END;
  IF Plot.yscale < 0.1 THEN
    Plot.yscale := 0.1;
  END;
END SetPlotScale;


PROCEDURE SetColourMap * (Plot             : PlotType;
                          ColourIndex      : INTEGER;
                          Red, Green, Blue : INTEGER);
(* Change the set of predefined colours for the plots *)
BEGIN
  (* g.SetRGB4(i.ViewPortAddress(Plot.Window),ColourIndex,Red,Green,Blue); *)
END SetColourMap;


PROCEDURE SetScatterPlot * (Plot  : PlotType);
(* Set to plot single points or a scatter plot *)
BEGIN
  Plot.ScatterPlot := TRUE;
  Plot.ScatterChar := "O";   (* default character *)
END SetScatterPlot;


PROCEDURE SetScatterChar * (Plot  : PlotType;
                            Ch    : CHAR);
(* Change the default `O' character for scatter plots *)
BEGIN
  Plot.ScatterChar[0] := Ch;
END SetScatterChar;


PROCEDURE InitOffsetPlot *
                  (VAR p                  : PlotType;
                   PlotKind               : PlotKindType;
                   xMin, xMax, yMin, yMax : LONGREAL;
                   TopBorder              : INTEGER;
                   width, height          : INTEGER;
                   xDiv, yDiv             : INTEGER;
                   xSubDiv, ySubDiv       : INTEGER;
                   xDec, yDec             : INTEGER;
                   xOffset, yOffset       : INTEGER;
                   OldPlot                : PlotType)
                   : BOOLEAN;
VAR
  Ok      : BOOLEAN;
  Border1 : INTEGER;
  Border2 : INTEGER;
  DumStr  : ARRAY 40 OF CHAR;
BEGIN
  NEW(p);
  IF p = NIL THEN
    RETURN FALSE;
  END;

  (* If OldPlot is defined use screen *)
  IF OldPlot # NIL THEN
    (* p.Window := OldPlot.Window;

    (* Extract the RastPort drawing surface information *)
    p.Drawing := OldPlot.Drawing;
    *)
  END;

  (* Initialize the plot variables for later use *)
  IF RealToString(yMax, DumStr, 2, yDec, " ") THEN
    Border1 := SHORT((str.Length(DumStr) + 3) * CharWidth)
  ELSE Border1:=0
  END;
  IF RealToString(yMin, DumStr, 2, yDec, " ") THEN
    Border2 := SHORT((str.Length(DumStr) + 3) * CharWidth)
  ELSE Border2:=0
  END;
  IF Border1 > Border2 THEN
    p.GMnx := Border1;
  ELSE
    p.GMnx := Border2;
  END;
  p.GMny := TopBorder + yOffset;
  p.GMxx := width - p.GMnx DIV 2 + xOffset;
  p.GMxy := height - CharHeight * 3 + yOffset;
  INC(p.GMnx, xOffset);
  p.Divx := xDiv;  p.SubDivx := xSubDiv; p.Decx := xDec;
  p.Divy := yDiv;  p.SubDivy := ySubDiv; p.Decy := yDec;
  p.Scrw := width; p.Scrh := height;
  SetPlotOffset(p, 0, 0);  (* no initial offset   *)

  (* No scaling initially *)
  p.xscale := 1.0;
  p.yscale := 1.0;

  (* Miscellaneous plot values *)
  p.PlotColour  := Colour1;     (* default plot colour *)
  p.TextColour  := Colour1;     (* default text colour *)
  p.GridColour  := Colour2;     (* default grid colour *)
  p.PlotIs      := PlotKind;
  p.GridDrawn   := FALSE;
  p.UserLabels  := FALSE;       (* use internal labels *)
  p.ScatterPlot := FALSE;       (* not a scatter plot  *)
  p.ZeroLineOn  := TRUE;        (* zero line enabled   *)
  p.CopyOfAPlot := TRUE;        (* this is a copy      *)

  (* Initialize the plot minimum and maximums *)
  SetPlotLimits(p, xMin, xMax, yMin, yMax);
  RETURN TRUE;
END InitOffsetPlot;


PROCEDURE InitPlot * (VAR Plot               : PlotType;
                      (* Window                 : i.WindowPtr; *)
                      PlotKind               : PlotKindType;
                      xMin, xMax, yMin, yMax : LONGREAL;
                      xDiv, yDiv             : INTEGER;
                      xSubDiv, ySubDiv       : INTEGER;
                      xDec, yDec             : INTEGER)
                      : BOOLEAN;
(* Set up the initial plot characteristics and create
   the initial custom graph screen. *)
VAR
  width, height, loff, toff : INTEGER;
BEGIN
  (* Initialize most of the plot variables *)
  (*
  width := Window.width DIV 2; height := Window.height-4;
  IF ~InitOffsetPlot(Plot, PlotKind,
                        xMin, xMax, yMin, yMax,
                        Window.borderTop+3,
                        width, height, xDiv, yDiv,
                        xSubDiv, ySubDiv,
                        xDec, yDec, 0, 7, NIL) THEN
    RETURN FALSE;
  END;

  (* Extract the RastPort drawing surface information *)
  Plot.Window := Window;
  Plot.Drawing := Window.rPort;
  *)
  Plot.CopyOfAPlot := FALSE; (* this is the original *)
  RETURN TRUE;
END InitPlot;


PROCEDURE DonePlot * (VAR Plot : PlotType);
(* Close off the plot -- when done *)
BEGIN
  IF Plot # NIL THEN
    Plot := NIL
  END;
END DonePlot;


PROCEDURE PlotZero(p : PlotType);
(* Plot the zero line *)
BEGIN
  (* emphasize y = 0 line *)
  IF p.ZeroLineOn THEN
    IF p.ScatterPlot THEN
      (* g.SetDrMd(p.Drawing, g.jam2); *)
      p.ScatterPlot := FALSE;
      IF ~(p.PlotIs IN YLog) THEN
        IntPlotLine(p, Colour3, p.Minx, 0.0, p.Maxx, 0.0);
      END;
      p.ScatterPlot := TRUE;
    ELSE
      IF ~(p.PlotIs IN YLog) THEN
        IntPlotLine(p, Colour3, p.Minx, 0.0, p.Maxx, 0.0);
      END;
    END;
  END;
END PlotZero;


PROCEDURE NewYLimits * (p : PlotType; Fx : PlotFunction);
VAR
  int, x, y : LONGREAL;
BEGIN
  (* make the window busy *)
  (* ok := bc.SetWindowBusy(p.Window); *)

  int := 5*(p.Maxx - p.Minx) / (p.GMxx - p.GMnx);

  (* go through the range of x values and plot y values *)
  x := p.Minx; p.Miny := 1.0E10; p.Maxy := -1.0E10;
  REPEAT
    CASE p.PlotIs OF
      LogX :
        y := Fx(TenTo(x));
        |
      LogY :
        y := LogOf(Fx(x));
        |
      Log :
        y := LogOf(Fx(TenTo(x)));
        |
      ELSE
        y := Fx(x);
    END;
    IF y < p.Miny THEN p.Miny := y END;
    IF y > p.Maxy THEN p.Maxy := y END;
    x := x + int
  UNTIL x > p.Maxx;
  IF p.Miny = p.Maxy THEN
    p.Maxy := p.Maxy + 0.5; p.Miny := p.Miny - 0.5
  END;

  (* clear busy indication *)
  (* bc.ClrWindowBusy(p.Window) *)
END NewYLimits;


PROCEDURE PlotFx * (p : PlotType; Fx : PlotFunction);
(* Plot the passed function *)
VAR
  lastx, lasty, x, y, int, dx, dy : LONGREAL;
  OldMinx, OldMaxx, OldMiny, OldMaxy : LONGREAL;
  maxDigits: INTEGER; ok : BOOLEAN;
BEGIN
  (* make the window busy *)
  (* ok := bc.SetWindowBusy(p.Window); *)

  (* only use 8 digits for faster plots *)
  maxDigits := X.GetMaxDigits();
  X.SetMaxDigits(8);

  (* save the old plot window dimensions *)
  OldMinx := p.Minx;  OldMaxx := p.Maxx;
  OldMiny := p.Miny;  OldMaxy := p.Maxy;

  (* calculate scaled plot window dimensions *)
  dx := (p.Maxx - p.Minx) / p.xscale;
  dy := (p.Maxy - p.Miny) / p.yscale;
  p.Minx := p.Minx + dx * (p.xscale - 1.0) * 0.5;
  p.Maxx := p.Minx + dx;
  p.Miny := p.Miny + dy * (p.yscale - 1.0) * 0.5;
  p.Maxy := p.Miny + dy;

  (* draw the plot grid *)
  DrawGrid(p);

  (* initialize the plot variables *)
  lastx := p.Minx;
  lasty := Fx(p.Minx);
  x := lastx;
  y := lasty;
  IF p.PlotIs IN {Bar, Line} THEN
    int := (p.Maxx - p.Minx) / p.Divx;
  ELSE
    int := 5*(p.Maxx - p.Minx) / (p.GMxx - p.GMnx);
  END;

  IF p.ScatterPlot THEN
    (* g.SetDrMd(p.Drawing, g.jam1); *)
  END;

  (* go through the range of x values and plot y values *)
  REPEAT
    IF p.PlotIs = Bar THEN
      PlotBar(p, p.PlotColour, int * 0.25, x, y);
    ELSE
      IntPlotLine(p, p.PlotColour, lastx, lasty, x, y);
    END;
    lastx := x;
    lasty := y;
    x := x + int;
    CASE p.PlotIs OF
      LogX :
        y := Fx(TenTo(x));
        |
      LogY :
        y := LogOf(Fx(x));
        |
      Log :
        y := LogOf(Fx(TenTo(x)));
        |
      ELSE
        y := Fx(x);
    END;
  UNTIL x > p.Maxx+int;
  PlotZero(p);

  (* restore the old plot window dimensions *)
  p.Minx := OldMinx;  p.Maxx := OldMaxx;
  p.Miny := OldMiny;  p.Maxy := OldMaxy;

  (* free busy window *)
  (* bc.ClrWindowBusy(p.Window); *)
  X.SetMaxDigits(maxDigits)
END PlotFx;


PROCEDURE SetPlotColour * (Plot   : PlotType;
                           Colour : INTEGER);
(* Change the plot colour *)
BEGIN
  Plot.PlotColour := Colour;
END SetPlotColour;


PROCEDURE SetTextColour * (Plot   : PlotType;
                           Colour : INTEGER);
(* Change the plot colour *)
BEGIN
  Plot.TextColour := Colour;
END SetTextColour;


PROCEDURE SetGridColour * (Plot   : PlotType;
                           Colour : INTEGER);
(* Change the plot colour *)
BEGIN
  Plot.GridColour := Colour;
END SetGridColour;


PROCEDURE SetZeroLine * (Plot : PlotType;
                         On   : BOOLEAN);
(* Enable/disable zero line *)
BEGIN
  Plot.ZeroLineOn := On;
END SetZeroLine;


PROCEDURE SetLabelRoutine * (Plot         : PlotType;
                             LabelRoutine : LabelProc);
(* Overrides the numerical labels with user-defined
   labels for each grid position along the x-axis *)
BEGIN
  Plot.GetLabel   := LabelRoutine;
  Plot.UserLabels := TRUE;          (* use LabelRoutine *)
END SetLabelRoutine;


PROCEDURE CenterLabelX * (p     : PlotType;
                          Label : ARRAY OF CHAR;
                          y     : INTEGER);
(* Center the text label horizontally *)
VAR
  x : INTEGER;
BEGIN
  x := (p.GMxx - p.GMnx) DIV 2 + p.GMnx;   (* center position *)
  DEC(x, str.Length(Label) * CharWidth DIV 2); (* offset *)
  LabelX(p, Label, x, y);
END CenterLabelX;


PROCEDURE CenterLabelY * (p     : PlotType;
                          Label : ARRAY OF CHAR;
                          x     : INTEGER);
(* Center the text label vertically *)
VAR
  y : INTEGER;
BEGIN
  y := (p.GMxy - p.GMny) DIV 2 + p.GMny;   (* center position *)
  DEC(y, str.Length(Label) * CharHeight DIV 2);
  LabelY(p, Label, x, y);
END CenterLabelY;


PROCEDURE LabelMinMax * (p  : PlotType; Fx : PlotFunction);
(* Label local minima and maxima by tracking whether the
   first derivative of Fx changes sign *)
CONST
  Positive = 0;
  Negative = 1;
TYPE
  SignType = SHORTINT;
VAR
  sign, signp : SignType;
  x, xlog, xplot, y, yp, yplot, delta : LONGREAL;
  Gx, Gy : INTEGER;
  s : ARRAY 40 OF CHAR;
BEGIN
  x := p.Minx;
  signp := Positive;
  IF p.PlotIs IN {Bar, Line} THEN
    delta := (p.Maxx - p.Minx) / p.Divx;
  ELSE
    delta := (p.Maxx - p.Minx) / (p.GMxx - p.GMnx);
  END;
  CASE p.PlotIs OF
    LogX :
      yp := Fx(TenTo(x - delta));
      |
    LogY :
      yp := LogOf(Fx(x - delta));
      |
    Log :
      yp := LogOf(TenTo(Fx(x - delta)));
      |
    ELSE
      yp := Fx(x - delta);
  END;
  REPEAT
    IF p.PlotIs IN XLog THEN
      xlog := TenTo(x);
    ELSE
      xlog := x;
    END;
    IF p.PlotIs IN YLog THEN
      y := LogOf(Fx(xlog));
    ELSE
      y := Fx(xlog);
    END;
    IF y - yp >= 0.0 THEN
      sign := Positive;
    ELSE
      sign := Negative;
    END;
    IF (sign # signp) OR
       ((p.PlotIs = Bar) & (yp # 0.0)) THEN
      (* found a local minima/maxima *)
      IF p.PlotIs IN {Bar, Line} THEN
        xplot := x - delta;
        yplot := yp;
      ELSE
        xplot := x;
        yplot := y;
      END;
      xyToCoords(p, xplot, yplot, Gx, Gy);
      IF yplot <= 0.0 THEN
        INC(Gy, CharHeight);  (* value below curve *)
      ELSE
        DEC(Gy, CharHeight DIV 3);  (* above curve *)
      END;
      IF RealToString(yplot, s, 2, p.Decy+2, " ") THEN
        DEC(Gx, str.Length(s) * CharWidth DIV 2);
        LabelX(p, s, Gx, Gy);
      END;
      signp := sign;
    END;
    x := x + delta;
    yp := y;
  UNTIL x > p.Maxx;
END LabelMinMax;


PROCEDURE InformationBox * (p             : PlotType;
                            Position      : PositionType;
                            Lines         : ARRAY OF LineType;
                            FillColour    : INTEGER;
                            OutlineColour : INTEGER);
(* Place an information box in the plot area *)
VAR
  Gx, Gy    : INTEGER;
  Gxs, Gys  : INTEGER;
  Gw, Gh    : INTEGER;
  ChCnt     : INTEGER;
  LineCnt   : INTEGER;
BEGIN
  Gxs := 2 * CharWidth;       (* side border width  *)
  Gys := 2 * CharHeight;      (* upper/lower border *)

  (* Determine the text box width and height *)
  Gh := 2 * Gys + SHORT(LEN(Lines)) * CharHeight;
  Gw := 2 * Gxs;
  ChCnt := 0;
  FOR LineCnt := 0 TO SHORT(LEN(Lines))-1 DO
    IF str.Length(Lines[LineCnt]) > ChCnt THEN
      ChCnt := SHORT(str.Length(Lines[LineCnt]));
    END;
  END;
  INC(Gw, ChCnt * CharWidth);

  (* Determine the text box position *)
  CASE Position OF
    UpperLeft :
      Gx := p.GMnx;
      Gy := p.GMny;
      |
    UpperMiddle :
      Gx := (p.GMxx - p.GMnx - Gw) DIV 2 + p.GMnx;
      Gy := p.GMny;
      |
    UpperRight :
      Gx := p.GMxx - Gw;
      Gy := p.GMny;
      |
    LowerLeft :
      Gx := p.GMnx;
      Gy := p.GMxy - Gh;
      |
    LowerMiddle :
      Gx := (p.GMxx - p.GMnx - Gw) DIV 2 + p.GMnx;
      Gy := p.GMxy - Gh;
      |
    LowerRight :
      Gx := p.GMxx - Gw;
      Gy := p.GMxy - Gh;
      |
  END;

  (*
  (* Draw the box *)
  g.SetAPen(p.Drawing, OutlineColour);
  g.RectFill(p.Drawing, Gx, Gy, Gx+Gw, Gy+Gh);
  g.SetAPen(p.Drawing, FillColour);
  g.RectFill(p.Drawing, Gx+2, Gy+2, Gx+Gw-2, Gy+Gh-2);

  (* Output the text strings *)
  INC(Gx, Gxs);
  INC(Gy, Gys);
  g.SetBPen(p.Drawing, FillColour); (* set text background *)
  FOR LineCnt := 0 TO SHORT(LEN(Lines))-1 DO
    INC(Gy, CharHeight);
    LabelX(p, Lines[LineCnt], Gx, Gy);
  END;
  g.SetBPen(p.Drawing, BackColour); (* normal background   *)
  *)
END InformationBox;

END UnixPlotLibrary.
